/*********************************************************/
/* axttyd.c: AXEL tty server                             */ 
/*********************************************************/
/* Version:                                              */
/*    - 1.1a: first version                              */
/*    - 1.1b: 64 association are allowed                 */
/*    - 1.1c:                                            */
/*      FIXED BUG: max file descriptor onto select       */
/*      instruction (64->128)                            */
/*      MODIFICATION: setpgrp() added at the beginning   */
/*      to detach the process from the caller            */
/*    - 1.1d:                                            */
/*      FIXED BUG: move the setpgrp instruction to really*/
/*      detach the process from a tty                    */
/*    - 1.2a:                                            */
/*      NEW: axttyd handles at the same time the ptty and*/
/*      the ptmx devices                                 */
/*      NEW: 2 SCO versions (SCO3 and SCO5)              */
/*      NEW: 2 SUN versions (SUN4 and SUN5)              */
/*    - 1.2b:                                            */
/*      NEW: axttyd allows connexion from sessX          */
/*      NEW: date and time are included in the logfile   */
/*      MODIF: MAX_ENTRY is now 128                      */
/*      MODIF: Do not perform ioctl for a second conn.   */
/*    - 1.2c:                                            */
/*      BUG: cast keep with (void *)                     */
/*      MODIF: change AX3000 by 'device'                 */
/*    - 1.2d - 16/03/98:                                 */
/*      MODIF: delete the test about the device keyword  */
/*      (for example: sess1, port1...)                   */
/*    - 1.2e - 14/04/98:                                 */
/*      MODIF: max buff is now 30 (instead of 50)        */
/*      NEW: DEC ALPHA runtime                           */
/*      MODIF: add the 'errno 5' test (SUN OS)           */
/*    - 1.3a - 11/05/98:                                 */
/*      NEW: Add the '-b' option. This option allows to  */
/*      use a buffer for the data from the socket.       */
/*      MODIF: the flag 'first' is now an array.         */
/*    - 1.3b TEST - 30/06/98:                            */
/*      NEW: A flush is performed on ptty before the     */
/*      main loop of a son (bug of select)               */
/*    - 1.3b - 07/07/98:                                 */
/*      Test version is certificated                     */
/*      axttyd acts also as the client part              */
/*    - 1.3b REV1:                                       */
/*      the flush code has been removed                  */
/*    - 1.3b REV2:                                       */
/*      the EINTR select error is now ignored            */
/*      the SIGUSR1 signal is now ignored by child       */
/*    - 1.3b REV3:                                       */
/*      Push the ldterm and ptem modules for stream ptty */
/*    - 1.3b REV4:                                       */
/*      Add IdConn function (no-idstring test)           */
/*      Add strtime global variable                      */
/*      Add LINUX option                                 */
/*    - 1.3c                                             */
/*      Add REDHAT option and remove LINUX one           */
/*    - 1.3c R1                                          */
/*      set MAX_ENTRY to 256                             */
/*    - 1.3c R2                                          */
/*      Add -F parameter to flush ptty when a connection */
/*      is opened.                                       */
/*      In the association file, a '#' can be used       */
/*      anywhere in a line                               */
/*      A negative value of WriteSocket is processed     */
/*    - 1.3c R3                                          */
/*      A close is performed after a shutdown.           */
/*      the 'maxfd' variable is now used for select      */
/*      The 'listen' return value is now checked.        */
/*      MAX_CONN is now 256                              */
/*    - 1.3c R4                                          */
/*      modif. declare of buffsock in FlushBuffSock      */
/*      ALPHA was defined two times.                     */
/*    - 1.3c R5                                          */
/*      UnixWare 7.1 version: include list has been modi */
/*      fied. clilen and my_lg types are size_t.         */
/*    - 1.3c R6                                          */
/*      ALPHA: slave tty line disc. set to raw mode      */
/*    - 1.3c R7                                          */
/*      LYNX version: setpgrp                            */
/*      File desc 0, 1, 2 are closed after init.         */
/*      Extra file descriptors are now closed by sons    */
/*      Include files have been added to fix warnings    */
/*    - 1.3c R8                                          */
/*      For AIX: using ptc instead of ptmx.              */
/*    - 1.3c R9                                          */
/*      When ptmx (or ptc) master is used, a symbolic    */
/*      name can be used for slave. This avoid to open   */
/*      many times the master to get the right slave     */
/*    - 1.4a (shipped also 1.3c R10)                     */
/*      The master pty is opened with O_NDELAY to prevent*/
/*      blocking write operation. When 0 byte is write   */
/*      on pty, an incremental timer is enbled and a     */
/*      message is logged.                               */
/*      The '-U' option is used for uni-directional      */
/*      communication: bytes from Axel device are not    */
/*      written on ptty (they are dropped)               */
/*      A chmod 666 is applied to /dev/ttypx             */
/*      After a connection attempt, we check is the ttyp */
/*      has been successfully opended.                   */
/*      Genereric LINUX version                          */
/*    - 1.4b                                             */
/*      PID of forked processes are displayed            */
/*    - 1.4c                                             */
/*      When a ERR 11 (EAGAIN) occures, the write        */
/*      operation is not aborted                         */
/*      O_NDELAY is only used with '-U' option           */
/*    - 1.4d                                             */
/*      Check Mode: this mode is only available for SCO. */
/*      It allows to drop data from the network when the */
/*      slave ttyp is not opened. This is useful for     */
/*      devices which sends data continuously (such as   */
/*      scales).                                         */
/*      HP: select.h is not needed to compile on HP9000  */
/*    - 1.4e                                             */
/*      Remove REDHAT option                             */
/*      For LINUX, include time.h instead of sys/time.h  */
/*    - 1.4f                                             */
/*      axttyd -k option allows to disable keepalive     */
/*    - 1.4g                                             */
/*      The number of seconds are now included in the    */
/*      date given in the log messages                   */
/*      For the select timer, the value of the timer is  */
/*      set each time the select instruction is done.    */
/*      (this is mandatory for Linux support)            */
/*    - 1.4g1                                            */
/*      Change keepalive setting method                  */
/*    - 1.4g2                                            */
/*      Two new axttyd parameters allows to tune the     */
/*      RTTY disconnection conditions:                   */
/*         -rmi nb: max number of incoming bytes         */
/*         -rt sec: incativity timer                     */
/*      The PushModule function is not defined for LINUX */
/*      For LINUX, include time.h                        */
/*    - 1.4g3                                            */
/*      The flush option (-F) is also efficient for an   */
/*      rtty connectionn (slave ptty flushed)            */
/*    - 1.5a                                             */
/*      Added parameters -ns and -nt options. If -ns is  */
/*      different than 0, after sending -ns characters   */
/*      there will be -nt msec delay to reduce the load  */
/*      of axttyd (When sending continously data)        */
/*    - 1.5b                                             */
/*      The setpgrp instructions are now call before the */
/*      ClientTTY function. It's allow to detach these   */
/*      processes also.                                  */
/*      The syntax // is not supported by AIX            */
/*    - 1.5c                                             */
/*      Error messages are now begun by the date and the */
/*      process id.                                      */
/*      A loop allows now the TCP port to be opened.     */
/*      The SIG_ALARM is now replaced by a simple sleep. */
/*    - 1.5d                                             */
/*      Ported for SCO6                                  */
/*    - 1.5e                                             */
/*      The FLUSH is now performed by a 'select' (on RH5 */
/*      the O_NONBLOCK option doesn't work properly)     */
/*********************************************************/

#if ! defined (SUN4) && ! defined (LINUX)  && ! defined (OLDLINUX) && ! defined (HP)
#include <sys/select.h>
#endif


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#if defined (ALPHA)  || defined (OLDLINUX)
#include <sys/time.h>
#else
#if defined (UW7)
#include <sys/time.h>
#else
#include <time.h>
#endif
#include <sys/stropts.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <fcntl.h>
#include <termio.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#ifndef SCO5
#undef CHECK_OPEN_SLAVE	/* this mode is only available for SCO */
#else
#include <prototypes.h>
#endif

#ifdef CHECK_OPEN_SLAVE
#undef PTMX             	/* only ttypx are handled by this mode */
#include <sys/user.h>
#include <sys/spt.h>
#include <filehdr.h>
#include <syms.h>
#endif

#ifdef ICL
struct timeval 
   {
   long tv_sec;
   long tv_usec;
   };
#endif

#ifdef ICL
#define OS		"ICL SV 4.2 MP 7MPLUS.9.8"
#endif

#ifdef UW7 
#define OS		"UNIXWARE 7.1"
#endif

#ifdef SCO5
#define OS		"SCO OPENSERVER"
#endif

#ifdef SCO6
#define OS              "SCO OPENSERVER 6"
#endif

#ifdef SCO3
#define OS		"SCO UNIX 3.2.x"
#endif

#ifdef SVR4
#define OS		"UNIX SVR4"
#endif

#ifdef SUN4
#define OS		"SUN OS 4.1"
#endif

#ifdef SUN5
#define OS		"SUN SOLARIS 5.5.1"
#endif

#ifdef AIX
#define OS		"IBM AIX"
#endif

#ifdef NCR
#define OS		"UNIX NCR"
#endif

#ifdef HP 
#define OS		"HP9000"
#endif

#ifdef DG 
#define OS		"DATA GENERAL - AViiON"
#endif

#ifdef SNX
#define OS		"SINIX - RM400"
#endif

#ifdef INTER
#define OS		"INTERACTIVE V4.1"
#endif

#ifdef ALPHA
#define OS		"Digital Unix V4.0"
#endif

#ifdef OLDLINUX
#define OS		"LEGACY LINUX"
#endif

#ifdef LINUX
#define OS		"LINUX"
#endif

#ifdef LYNX  
#define OS		"LYNX Vxxx"
#endif


#define VERSION    "1.5e"

#define RTTYTIMER  60  
#define RTTYMAXIN  0  
#define TCPPORT    2048
#define LOGFILE    "/tmp/axttylog"
#define AXFILE     "/etc/axfile"

#define MAX_CONN   256		/* for listen instruction */
#define MAXBUFSIZE 40
#define MAX_LINE   80
#define MAX_ENTRY  256		/* for internal arrays */

#define MAX_FD     256          /* for PTMX use */

#define MAXBUFFSOCK   384
#define HIGHWATERMARK 280

#define NULLTIMER (struct timeval *)0

typedef char t_tab[MAX_ENTRY][4][30];

int Counter;
int NapSize;
int NapTime;

int nb_ax;
t_tab axtab;
int mfd[MAX_ENTRY];
int sfd[MAX_ENTRY];
pid_t axpid[MAX_ENTRY];
int mark[MAX_ENTRY];
int first[MAX_ENTRY];
int tcpport[MAX_ENTRY];
char linkname[MAX_ENTRY][30];
int nb_tty_clt;

int nb_new;
t_tab axnew;
int newmfd[MAX_ENTRY];
int newsfd[MAX_ENTRY];
pid_t newpid[MAX_ENTRY];
int newfirst[MAX_ENTRY];

char ax_file[MAX_LINE];
FILE *fd_log;

int use_buff;
int use_keepalive;
int do_flush;
int do_unidirectional;
int father_pid;
char strtime[MAX_LINE];
int rtty_timer;
int rtty_maxchar_in;

#ifdef CHECK_OPEN_SLAVE
int fdmem = -1; /* these 3 variables are used by each son (fork) */
long place;
int numptty;
int previous_state;
struct mynlist {
   char*           m_name;
   long            m_value;
   short           m_scnum;
   unsigned short  m_type;
   char            m_sclass;
   char            m_numaux;
   };

char* string_table;
int nsyms;
long  symptr;
FILE* fdnlist;

#define MPFL      nl[0].m_value

struct mynlist nl[] = {  
                      { "mptflags", 0 },
                      { 0 },
                    };

read_string_table ()
{
long strings;
long strlength;

strings = symptr + (SYMESZ * nsyms);
fseek (fdnlist, strings, SEEK_SET);
fread (&strlength, 4 ,1, fdnlist);
if (strlength)
   {
   string_table = (char *)malloc (strlength);
   fseek (fdnlist, strings, SEEK_SET);
   fread (string_table, strlength, 1, fdnlist);
   }
}

eval_sym (SYMENT*  se, struct mynlist ml[], int nsrch, int* sizelst)

{
int x, found = 0;

for (x=0;x<nsrch;x++)
   {
   if (se->n_zeroes)
      {
      if (strncmp(ml[x].m_name, se->n_name, 8) == 0)
         found=1;
      }
      else
      {
      if (strcmp(ml[x].m_name, &string_table[se->n_offset]) == 0)
         found=1;
      }
   if (found)
      {
      ml[x].m_value = se->n_value;
      ml[x].m_scnum = se->n_scnum;
      ml[x].m_type  = se->n_type; 
      ml[x].m_sclass= se->n_sclass;
      ml[x].m_numaux= se->n_numaux;
      found = 0;
      *sizelst--;
      }
   }
}


mynlist (char *obj_file, struct mynlist ml[])
{
struct filehdr  f_hdr;
SYMENT   se;
int x, ncs=0, sizelst=0, nsrch=0;

fdnlist = fopen (obj_file, "r");

while (ml[sizelst].m_name != '\0' )
   {
   sizelst++;
   nsrch++;
   }

fread (&f_hdr, sizeof(FILHDR), 1 , fdnlist);
nsyms = f_hdr.f_nsyms;
symptr = f_hdr.f_symptr;

printf("Nsyms : %d\n", nsyms);
read_string_table();

while ((ncs++ != nsyms) || (sizelst == 0))
   {
   fseek (fdnlist, symptr, SEEK_SET);
   fread (&se, SYMESZ, 1 ,fdnlist);
   symptr += SYMESZ;
   eval_sym (&se, ml, nsrch, &sizelst);
   for (x=0; x<se.n_numaux; x++)
      symptr += SYMESZ;
   }
}
#endif    /* CHECK_OPEN_SLAVE */


/*********************************************************/
/* This function is called by the SIGCLD signal.         */ 
/* The father cleans the association array entry of the  */
/* died child.                                           */
/*********************************************************/
char *GetTime ()
{
time_t clock;
struct tm *mydate;

time(&clock);
mydate=localtime(&clock);

sprintf(strtime,"%.2d/%.2d/%d-%.2d:%.2d:%.2d - ", mydate->tm_mday, mydate->tm_mon+1,mydate->tm_year+1900,mydate->tm_hour,mydate->tm_min, mydate->tm_sec);
return (strtime);
}

/*********************************************************/
/* This function is called by the SIGCLD signal.         */ 
/* The father cleans the association array entry of the  */
/* died child.                                           */
/*********************************************************/
void EndChild ()
{
int status;
void (*func)();
int p, dead, i;

p = waitpid (-1, &status, WNOHANG);	/* get the pid child */
dead = -1;
for (i=0;i<nb_ax;i++)			/* try to identify the child */
   if (p == axpid[i])
      {
      dead = i;				/* OK, I get it */
      break;
      }
fprintf(fd_log, "%s", GetTime());
if (dead == -1)
   fprintf(fd_log, "WARNING: can not identified a dead child.\n");
   else
   {
   axpid[dead] = -1;
   fprintf(fd_log, "Connection from '%s %s': closed.\n", axtab[dead][0], axtab[dead][1]);
   }
func = EndChild;			/* I re-armed the signal catcher */
signal (SIGCLD, func);
}

/*********************************************************/
/* This function is called by the PerformConn function.  */
/*********************************************************/
int FlushBuffSock(buffsock, ptrbuffsock, pttyfd, pttyname)

unsigned char *buffsock;
int *ptrbuffsock;
int pttyfd;
char *pttyname;
{
int l, p, w;
int i;
int nbsec;
#ifdef CHECK_OPEN_SLAVE
char mpstat;
#endif

p = 0;
nbsec=1;
l=(*ptrbuffsock);
if (do_unidirectional)
   {
   fprintf(fd_log, "%sUni-directional mode: %d bytes for %s are dropped\n",GetTime(),l-p, pttyname);
   fflush (fd_log);
   (*ptrbuffsock)=0;
   return (1);   
   }

#ifdef CHECK_OPEN_SLAVE
lseek (fdmem, place, SEEK_SET);
read (fdmem, &mpstat, sizeof (char));
   
if (!(mpstat & MPT_SPT_ASSOC))
   {    /* no slave associated, data is dropped */ 
   (*ptrbuffsock)=0;
   return (1);   
   }
/* else date is written to pttyfd */
#endif

while (p < l)	/* I must write exactely the r characters */
   {
   w = write (pttyfd, &buffsock[p], l-p);
   if (w == 0) 
      {
      fprintf(fd_log, "%sCan not write on %s: retry in %d sec.\n",GetTime(), pttyname, nbsec);
      fflush (fd_log);
      sleep (nbsec);
      if (nbsec < 4)
         nbsec *= 2;
      continue;
      }
   if (w < 0) 
      {
      if (errno != EAGAIN)
         {
         fprintf(fd_log, "%sWrite ERROR (%d) on %s: %d byte(s) lost\n",GetTime(),errno, pttyname,errno);
         fflush (fd_log);
         break;
         }
      sleep (nbsec);
      if (nbsec < 4)
         nbsec *= 2;
      continue;
      }
   if (nbsec > 1)
      {
      fprintf(fd_log, "%sOK, %s is now available\n",GetTime(), pttyname);
      fflush (fd_log);
      nbsec = 1;
      }
   p += w;
   }
(*ptrbuffsock)=0;
}

int WriteSocket (buf, r, fd)
char *buf;
int r;
int fd;
{
int p, w;

p = 0;
while (p < r)	/* I must write exactely the r characters */
   {
   w = write (fd, &buf[p], r-p);
   if (w < 0) 
      {
      fprintf(fd_log, "%schild process %d: can not write %d byte(s) on socket : %d\n",GetTime(), getpid(), r-p,errno);
      return (-1);
      }
   p += w;
   }
return (p);
}

/*********************************************************/
/* This is the main-loop child                           */
/* The child reads data from socket and writes it on ptty*/
/* The child reads data from ptty and writes it on socket*/
/*********************************************************/
int PerformConn (sockfd, pttyfd, first, pttyname)

int sockfd;			
int pttyfd;
int first;
char *pttyname;
{
int r, s;
char buf[MAXBUFSIZE];
struct termio new_tty;
fd_set f_r;
unsigned char buffsock[MAXBUFFSOCK];
int  ptrbuffsock;
struct timeval timer;
struct timeval *ptrtimer;
int maxfd;

#ifdef CHECK_OPEN_SLAVE

if (mynlist("/unix", nl) == -1)
   {
   fprintf(fd_log, "%sCHECK MODE: Can't get nlist for %\n", GetTime(), pttyname);
   fflush (fd_log);
   }
   else
   {
   fdmem=open ("/dev/kmem", O_RDONLY);
   if (fdmem < 0)
      {
      fprintf(fd_log, "%sCHECK MODE: Can't get kmem for %s\n", GetTime(), pttyname);
      fflush (fd_log);
      }
   numptty = atoi (&pttyname[9]);       
   place=MPFL+numptty*sizeof(char);
   previous_state = MPT_OPEN | MPT_BLK_FLAG;
   fprintf(fd_log, "%sCHECK MODE: OK, ptty #%d info located\n", GetTime(), numptty);
   fflush (fd_log);
   }
#endif

if (do_flush)
   { /** FLUSH **/    
   int tot;

   fprintf(fd_log, "%sFlush on %s in progress...\n", GetTime(), pttyname);
   fflush (fd_log);
   tot=0;
   while (1)
      {
      timer.tv_sec=0;
      timer.tv_usec=20000;
      ptrtimer=&timer;
      memset (&f_r, 0, sizeof(fd_set));
      FD_SET(pttyfd,&f_r);	
      s = select(pttyfd+1, &f_r, NULL, NULL, ptrtimer); 
      if (s == 0) break;
      if (FD_ISSET(pttyfd, &f_r))
         {
         r = read (pttyfd, buf, MAXBUFSIZE);
         if (r < 1) break;
         WriteSocket (buf, r, sockfd);
         tot+=r;
         }
      }
   if (tot > 0) 
      {
      fprintf(fd_log, "%sFLUSH: %d byte(s) from %s sent to the network\n", GetTime(), tot, pttyname);
      fflush (fd_log);
      }
   }

if (first)  /* this is the first time I accessed this ptty */
   { 
   ioctl(pttyfd,TCGETA,&new_tty);		/* load tty parameters */ 
   new_tty.c_iflag &= ~(ICRNL);
   new_tty.c_lflag &= ~(ICANON | ISIG | ECHO | ECHOE | ECHOK);
   new_tty.c_oflag &= ~(OPOST | ONLCR);
   new_tty.c_cc[VMIN] = 1;
   new_tty.c_cc[VTIME] = 1;
   ioctl(pttyfd,TCSETA,&new_tty); 		/* set new parameters */
   }

/* main loop */
ptrbuffsock=0;
Counter = 0;
ptrtimer=NULLTIMER;
maxfd=sockfd;
if (pttyfd>maxfd)
   maxfd=pttyfd;
while (1)
   {
   memset (&f_r, 0, sizeof(fd_set));
   FD_SET(sockfd,&f_r);		
   FD_SET(pttyfd,&f_r);	
   s = select(maxfd+1, &f_r, NULL, NULL, ptrtimer); 
   if (s == 0) 
      {   /* timer expired */
      FlushBuffSock(buffsock, &ptrbuffsock, pttyfd, pttyname);
      ptrtimer=NULLTIMER;
      }
   if (s < 0) 
      {
      if (errno == EINTR)
         continue;  /* it's not a real error, is it a signal */
      fprintf(fd_log, "%schild process %d: select error %d (s = %d)\n",GetTime(), getpid(),errno,s);
      fflush (fd_log);
      break;
      }
   if (FD_ISSET(pttyfd, &f_r))
      {					/* car received from ptty */
      int p, w;

      r = read(pttyfd, buf, 30); 
      if (r < 0) 
         {
         fprintf(fd_log, "%schild process %d: read from ptty error: %d\n",GetTime(), getpid(), errno);
         if (errno != 5)  /* Specific processing for SUN OS */
            break;
   	    else 
            {
            fprintf(fd_log, "  --> Give a new chance, sleeping during 5 seconds.\n");
            fflush (fd_log);
	    sleep (5);
	    continue;
            }
         }
      if (r == 0) 
         {
#ifdef CHECK_OPEN_SLAVE
         fprintf(fd_log, "%sCHECK MODE: child closes then re-open %s\n", GetTime(), pttyname);
         fflush (fd_log);
         close (pttyfd);
         pttyfd = OpenPTTY (pttyname);
         if (pttyfd < 0)
            exit (1);
#else 
         fprintf(fd_log, "%schild process %d: read no byte from ptty\n", GetTime(), getpid());
         fflush (fd_log);
#endif
         continue;
         }
      if (NapSize)
         {
         Counter = Counter + r;
         if (Counter > NapSize) 
            {
#if defined (SCO5) || defined (SCO3)
            nap(NapTime);
#else
            usleep(NapTime);
#endif
            Counter = 0;
            } 
         } 
      if (WriteSocket (buf, r, sockfd) <0) 
         break;   /* en this connection */
      }
   if (FD_ISSET(sockfd, &f_r))
      {					/* car received from socket */
      int p, w;

      r = read(sockfd, &buffsock[ptrbuffsock], 30); 
      if (r < 0) 
         {
         if (errno == ETIMEDOUT)
            fprintf(fd_log, "%schild process %d: the device is down.\n", GetTime(), getpid());
            else
            fprintf(fd_log, "%schild process %d: read from socket error: %d\n",GetTime(), getpid(), errno);
         fflush (fd_log);
         FlushBuffSock(buffsock, &ptrbuffsock, pttyfd, pttyname);
         ptrtimer=NULLTIMER;
         break;
         }
      if (r == 0) 
         {
         fprintf(fd_log, "%schild process %d: read no byte from socket\n", GetTime(), getpid());
         fflush (fd_log);
         FlushBuffSock(buffsock, &ptrbuffsock, pttyfd, pttyname);
         ptrtimer=NULLTIMER;
         break;
         }
      ptrbuffsock+=r;
      if ((use_buff == 0) || (ptrbuffsock >= HIGHWATERMARK))
         {
         FlushBuffSock(buffsock, &ptrbuffsock, pttyfd, pttyname);
         ptrtimer=NULLTIMER;
         }
         else
         {
         timer.tv_sec=0;
         timer.tv_usec=200000;
         ptrtimer=&timer;
         }
      }
   }
fprintf(fd_log, "%sChild process %d ended.\n\n", GetTime(),getpid());
fflush (fd_log);
close (pttyfd);
shutdown (sockfd, 2);
close (sockfd);
exit (1);
}
 

/*********************************************************/
/* This function is called by the CleanAll function.     */
/*********************************************************/
EndPrg (str)

char *str;
{
fprintf (fd_log, "%s%s\n",GetTime(), str);
fprintf (fd_log, "%sAXEL server ended.\n", GetTime());
exit (1);
}


/*********************************************************/
/* This function is called by the SIGHUP or SIGTERM sig. */
/*********************************************************/
void CleanAll ()
{
int i, mypid;

signal (SIGCLD, SIG_IGN);	/* now do not catch this signal */
mypid=getpid ();
fprintf (fd_log, "%sThe process (pid : %d) is killed by signal SIGCLD!\n", GetTime (), mypid);
if (mypid == father_pid)
   {
   for (i=0;i<nb_ax;i++)
      {
      if (axpid[i] != -1)
         {
         fprintf (fd_log, "%schild process %d is killed by a signal 9.\n", GetTime (), axpid[i]);
         kill (axpid[i], 9);	/* kill all the children */
         }
      if (linkname[i][0] != 0)
         unlink (linkname[i]);
      }
   fprintf (fd_log, "Axel server ended");
   }
fflush (fd_log);
exit (1);
}


/*********************************************************/
/* This function is called by ParseFile function         */
/* The character string return is a valid line (no       */
/* commentary and no byte lower than 32).                */
/*********************************************************/
GetLine (fd, str)

FILE *fd;
char *str;
{
int i, lg;
char *ptr;

while (1)
   {
   if (fgets (str, MAX_LINE, fd) == NULL)
      return (0);
   if (str[0] == '#')
      continue;				/* skip comments */
   ptr = strchr (str, '#');
   if (ptr != NULL) 
      *ptr = 0;                         /* skip char after comment */ 

   lg = strlen (str);
   str[lg] = 0;
   for (i=0;i<lg;i++)
      if ((str[i] == 0x0D) || (str[i] == 0x0A))
         {
         str[i] = 0;
         break;
         }
   break;				/* this line is valid */
   }
return (1);
}


/*********************************************************/
/* This function is called by ParseLine function         */
/* In the current string, the first word is isolated.    */
/*********************************************************/
char *FindToken (str)

char *str;
{
char *p;

p = str;
while ((*p))	/* while there is a non-null character */
   {
   if (((*p) == ' ') || ((*p) == 0x09))
      {		/* a separator is find */
      (*p) = 0;	/* the current word is terminated */
      p++;
      break;
      }
   p++;		 /* I add a valid character to the current word */
   }

while ((*p))	/* I skip the blank characters */
   if (((*p) == ' ') || ((*p) == 0x09))
      p++;
      else
      break;
return (p);
}


/*********************************************************/
/* This function is called by ParseFile function         */
/* The current line is examine to get each word and to   */
/* check the syntax (I need 4 word into the line).       */
/*********************************************************/
ParseLine (str, tab, n)

char *str;
t_tab tab;
int n;
{
char *ptr;
char *src;
int c;
char line [MAX_LINE];
char token [MAX_LINE];
char sep [5];
int i;

strcpy (line, str);
c = 0;
src = line;
while (1)
   {
   ptr = FindToken (src);		/* get an entire word */
   if ((*src) == 0)
      break;				/* no more word */
   strcpy (&tab[n][c][0], src);
   c++;
   src = ptr;
   }
if (c != 4)			/* A valid line is composed by 4 words */
   return (0);

return (1);
}

/*********************************************************/
/* This function tries to open slave and master ptty     */
/*********************************************************/
int OpenPTTY (master)

char *master;
{
int fd;
int oflag;

oflag = O_RDWR;
if (do_unidirectional == 1)
   oflag |= O_NDELAY;     /* to prevent blocking write   */
if (do_flush == 1)
   oflag |= O_NONBLOCK;   /* to perform a non blocking read instruction */

fd=open (master, oflag);
if (fd < 0)
   fprintf(fd_log, "%sCan not open %s (error %d).\n", GetTime(), master, errno);
return (fd);	
}


#ifdef PTMX
/*********************************************************/
/* This function tries to open slave and master ptty     */
/*********************************************************/
int OpenPTMX (master, slave, dolink)

char *master;
char *slave;
int dolink;
{
int i,j;
int fd[MAX_FD];
char my_slave[30];

for (i = 0; i < MAX_FD; i++)
   {
   fd[i] = open(master, O_RDWR);
   if (fd[i] < 0)
      break;			/* no more pts */
#ifndef AIX
   unlockpt (fd[i]);		/* unlock slave */
   strcpy (my_slave, (char *)ptsname (fd[i]));	/* slave's name */
#else
   strcpy (my_slave, (char *)ttyname (fd[i]));	/* slave's name */
#endif
   if (strcmp(slave, my_slave) == 0)
      break;			/* I found the good one */
   if (dolink)
      break;			/* exit on the first one */
   }
for (j = 0;j < i; j++)		/* for each master open descriptor */
   if (fd[j] > 0)			/* and valid descriptor */
      close (fd[j]);		/* I close it */
if ((fd[i] < 0) || (i == MAX_FD))
   {
   fprintf(fd_log, "%sERROR : Can't link %s and %s\n", GetTime(), master, slave);
   return (-1);
   }
strcpy (slave, my_slave);     /* in case of link */
return (fd[i]);
}
#else
/*********************************************************/
/* Dummy function for O.S. which doesn't support PTMX    */
/*********************************************************/
int OpenPTMX (master, slave, dolink)

char *master;
char *slave;
int dolink;
{
fprintf(fd_log, "%sERROR : PTMX devices not supported\n", GetTime());
return (-1);
}
#endif

#ifndef OLDLINUX
/*********************************************************/
/* Push the both ldterm and ptem modules for stream ptty */
/*********************************************************/
int PushModules(i, slave)

int i;
char *slave;
{
#ifdef PTMX
if (ioctl (sfd[i], I_PUSH, "ptem") < 0)
   {
   fprintf(fd_log, "   Can not push ptem on %s (error %d).\n", slave, errno);
   return (-1);
   }
if (ioctl (sfd[i], I_PUSH, "ldterm") < 0)
   {
   fprintf(fd_log, "   Can not push ldterm on %s (error %d).\n", slave, errno);
   return (-1);
   }
return (0);
#else
return (-1);
#endif
}
#endif

/*********************************************************/
/* This function use the association array to open the   */
/* ptyp and ttyp devices.                                */
/* This function is called after the ParseFile function  */
/* or the ReExamineFile function.                        */
/*********************************************************/
void AllocatePtty ()

{
int cpt, i;
char master[30];
char slave[30];
int module;
struct termio slv_tty;
int dolink;
int ret;

fprintf(fd_log, "\n%sDevice allocation:\n", GetTime());
fflush (fd_log);
cpt = 0;
for (i=0;i<nb_ax;i++)
   {			/* for every valid entry */
   module=0;
   dolink=0;
   linkname[i][0] = 0;
   strcpy (master, axtab[i][2]);
   strcpy (slave, axtab[i][3]);
   if (mfd[i] == -1)
      { /* not already opened */
#ifdef AIX
      if (strcmp (master, "/dev/ptc") == 0)
#else
      if (strcmp (master, "/dev/ptmx") == 0)
#endif
         {
         if (strncmp (slave, "/dev/pts", 8) != 0)
            {
            dolink=1;
            strcpy (linkname[i], slave);
            }
         mfd[i] = OpenPTMX (master, slave, dolink);
         }
         else
         mfd[i] = OpenPTTY (master);
      if (mfd[i] == -1)
         continue;
#ifdef CHECK_OPEN_SLAVE
      /* for this mode, axttyd opens only the master to check further if */
      /* a slave is associated */
      sfd[i] = -1;
#else
      sfd[i] = open (slave, O_RDWR);
      if (sfd[i] < 0)
         {
         fprintf(fd_log, "   Can not open %s (error %d).\n", slave, errno);
         close (mfd[i]);
         mfd[i]=-1; /* I can't open the slave so I close the master */
         continue;	/* ERR, next entry */
         }
#endif
      if (strcmp (master, "/dev/ptmx") == 0)
         {
#if ! defined (LINUX)  && ! defined (OLDLINUX) 
         if (PushModules (i, slave) < 0)
            {
            close (sfd[i]);
            sfd[i] =-1;
            close (mfd[i]);
            mfd[i] =-1;
            continue;
            }
         module=1;
#endif
         }
      }
   fprintf(fd_log, "   %s <-> %s", master, slave);
   chmod(slave,0666);
   if (dolink)
      {
      unlink(linkname[i]);   /* maybe it is needed */
      symlink(slave, linkname[i]);
      chmod(linkname[i],0666);
      fprintf(fd_log, " <-> %s", linkname[i]);
      }
   if (module)
      fprintf(fd_log, " (ptem et ldterm pushed)");
#ifdef ALPHA
      else
      {
      ioctl(sfd[i],TCGETA,&slv_tty);               /* load tty parameters */
      slv_tty.c_iflag &= ~(ICRNL | IXON);
      slv_tty.c_lflag &= ~(ICANON | ISIG | ECHO | ECHOE | ECHOK);
      slv_tty.c_oflag &= ~(OPOST | ONLCR);
      slv_tty.c_cc[VMIN] = 1;
      slv_tty.c_cc[VTIME] = 1;
      ioctl(sfd[i],TCSETA,&slv_tty);               /* set  tty parameters */
      fprintf(fd_log, " (slave tty set to raw mode)");
      }
#endif
   fprintf(fd_log, ": OK.\n");
   fflush (fd_log);
   cpt++;
   }
if (cpt == 0)
   fprintf(fd_log, "%sWarning: no device allocated (no association)!\n", GetTime());
fflush (fd_log);
}


/*********************************************************/
/* Examine for the association file.                     */
/* This function is called by main or by ReExaminFile.   */
/*********************************************************/
int ParseFile (tab, file)

t_tab tab;
char *file;
{
FILE *f;
struct hostent *host;
struct in_addr *p;
int nb;
char str[MAX_LINE];

if ((f = fopen (file, "r")) == NULL)
   {
   sprintf(str, "Can not open %s!\n", file);
   EndPrg (str);
   }
nb = 0;

nb_tty_clt = 0;
fprintf(fd_log, "\n%sAssociation File (%s):\n", GetTime (), file);
while (GetLine (f, str))		/* while there is a valid line */
   {
   if (!ParseLine (str, tab, nb))	/* examine this line */
      {
      fprintf(fd_log, "   %s... : syntax error.\n", tab[nb][0]);
      continue;
      }
   host = gethostbyname (tab[nb][0]);	/* check the host name */
   if (host == NULL)
      {
      errno = 0;
      fprintf(fd_log, "   %s : unknown device.\n", tab[nb][0]);
      continue;
      }

   fprintf(fd_log, "   %s: '%s' port associated with %s (%s)\n", tab[nb][0], tab[nb][1], tab[nb][2], tab[nb][3]);
   tcpport[nb]=atoi (tab[nb][1]);  /* 0 if it is not a real number */
   if (tcpport[nb] > 0)
      nb_tty_clt++;
   p = (struct in_addr *)host->h_addr_list[0];
   /* change the host name into IP address */
   strcpy (tab[nb][0], (char *)inet_ntoa(*p));
   nb++;			/* one more entry */
   }
fclose (f);

if (nb == 0)
   fprintf(fd_log, "%sWarning: no association!\n", GetTime());

fflush (fd_log);
return (nb);
}


/*********************************************************/
/* This function is called by the SIGUSR1 signal.        */
/* Another association array is filled then this array   */
/* and the 'real' association array are compared and the */
/* 'real' association is updated                         */
/*********************************************************/
void ReExamineFile()
{
void (*func) ();
int i,j,k;

signal (SIGUSR1, SIG_IGN);
nb_new = ParseFile (axnew, ax_file);	/* fill a new associatino array */
for (i=0;i<nb_new;i++)		/* compare new and real association array */
   {
   newpid[i] = -1;
   newmfd[i] = -1;
   newsfd[i] = -1;
   k = -1;
   /* try to find the same (AX3000/PORT) */
   for (j=0;j<nb_ax;j++)
      if ((strcmp(axnew[i][0], axtab[j][0]) == 0) && (strcmp(axnew[i][1], axtab[j][1]) == 0))
         {
         k = j; /* founded */
         break;
         }
   if (k != -1)
      { /* same AX3000/PORT */
      if ((strcmp(axnew[i][2], axtab[k][2]) == 0) && (strcmp(axnew[i][3], axtab[k][3]) == 0))
         { /* same ptty */
         mark [k] = 1;
         newpid[i] = axpid[k];
         newmfd[i] = mfd[k];
         newsfd[i] = sfd[k];
         newfirst[i] = first[k];
         continue;
         }
         else
         { /* different ptty */
         mark [k] = 1;
         if (axpid[k] != -1)
            kill (axpid[k], 9);
         continue;
         }
      } 
   }
/* I kill the non-marked old association */
for (j=0;j<nb_ax;j++)
   if (mark[j] == 0)
      {
      if (axpid[j] != -1)
         kill (axpid[j], 9);
      }
/* I copy newtab -> axtab */
nb_tty_clt = 0;
for (i=0;i<nb_new;i++)
   {
   strcpy (axtab[i][0], axnew[i][0]);
   strcpy (axtab[i][1], axnew[i][1]);
   strcpy (axtab[i][2], axnew[i][2]);
   strcpy (axtab[i][3], axnew[i][3]);
   axpid[i] = newpid[i];
   mfd[i] = newmfd[i];
   sfd[i] = newsfd[i];
   first[i] = newfirst[i];
   mark[i] = 0;
   }
nb_ax = nb_new;
AllocatePtty ();
func = ReExamineFile;
signal (SIGUSR1, func);			/* signal re-armed */
}


int OpenSocket (axname, tcpport)

char *axname;
int tcpport;
{
struct sockaddr_in server;
struct hostent *hp;   
int fd_sock;

server.sin_family = AF_INET;
hp = gethostbyname(axname);
if (hp == NULL)
   {
   fprintf(stderr, "ERROR: %s is an unknown AXEL device\n", axname);
   fflush (fd_log);
   return (-1);
   }
errno=0;		/* to fix the trouble of gethostbyname */
memcpy((char *) &server.sin_addr, (char *) hp->h_addr, hp->h_length);
server.sin_port = htons(tcpport);

fd_sock = socket(AF_INET, SOCK_STREAM, 0);
if (fd_sock < 0)
   {
   fprintf(stderr, "ERROR: can't create a socket !\n");
   fflush (fd_log);
   return (-1);
   }
#ifdef ICL
if (connect(fd_sock, (caddr_t)&server, sizeof server) < 0)
#else
if (connect(fd_sock, (struct sockaddr *)&server, sizeof server) < 0)
#endif
   {
   close(fd_sock);
   fprintf(fd_log, "Connection to %s/%d refused: ", axname,tcpport);
   switch (errno)
      { 
      case ECONNREFUSED:
         fprintf(fd_log, "tty server busy (or not started) or bad TCP Port Number\n");
         break;
      case ETIMEDOUT:
         fprintf(fd_log, "tty server does not respond...\n");
         break;
      case EHOSTDOWN:
         fprintf(fd_log, "The AXEL device is down...\n");
         break;
      default:
         fprintf(fd_log, "ERROR %d\n", errno);
         break;
      }
   fflush (fd_log);
   return (-1);
   }
fprintf(fd_log, "%sConnection with %s/%d established\n", GetTime(), axname, tcpport);
fflush (fd_log);
return (fd_sock);
}


int DoClientTTY (axname, tcpport, pttyfd, ttyname, slavefd)
char *axname;
int tcpport;
int pttyfd;
char *ttyname;
int slavefd;
{
int w, r, s;
unsigned char buf[MAXBUFSIZE];
struct termio new_tty;
fd_set f_r;
struct timeval timer;
struct timeval *ptrtimer;
int sockfd;
unsigned int totalin, totalout;
    
ioctl(pttyfd,TCGETA,&new_tty);		/* load tty parameters */ 
new_tty.c_iflag &= ~(ICRNL);
new_tty.c_lflag &= ~(ICANON | ISIG | ECHO | ECHOE | ECHOK);
new_tty.c_oflag &= ~(OPOST | ONLCR);
new_tty.c_cc[VMIN] = 1;
new_tty.c_cc[VTIME] = 1;
ioctl(pttyfd,TCSETA,&new_tty); 		/* set new parameters */


while (1)
   {
   ptrtimer=NULLTIMER;

   /* WAITING FOR 1 CHAR. */
   memset (&f_r, 0, sizeof(fd_set));
   FD_SET(pttyfd,&f_r);
   s = select(128, &f_r, NULL, NULL, ptrtimer); 
   if (s < 0) 
      {
      if (errno == EINTR)
         continue;  /* it's not a real error, is it a signal */
      fprintf(fd_log, "%schild process %d: select error %d (s = %d)\n",GetTime(), getpid(), errno,s);
      fflush (fd_log);
      sleep (10);
      continue;
      }
   /* FIRST CHAR RECEIVED */
   r = read(pttyfd, buf, 1); 
   if (r < 0) 
      {
      fprintf(fd_log, "%schild process %d: Can not read first char from %s (err %d) \n",GetTime (), getpid(), ttyname, errno);
      fflush (fd_log);
      continue;
      }
   fprintf(fd_log, "%schild process %d: First char received from %s\n", GetTime(), getpid(), ttyname);
   fflush (fd_log);

   /* OPEN SOCKET */
   while (1)
      {
      sockfd = OpenSocket (axname, tcpport);
      if (sockfd < 0) 
         sleep (10);
         else
         break;
      }

   /* WRITING FIRST CHAR */
   w = write (sockfd, buf, 1); 
   if (w < 0) 
      {
      fprintf(fd_log, "%schile process %d: can not write first char to %s (err %d)\n",GetTime (), getpid(), ttyname, errno);
      return (-1);
      }

   /* MAIN LOOP */
   totalin=0;
   totalout=1;
   while (1)
      {
      if (rtty_timer == 0)
         ptrtimer=NULLTIMER;
         else
         {
         timer.tv_sec=rtty_timer;
         timer.tv_usec=0;
         ptrtimer=&timer;
         }
      memset (&f_r, 0, sizeof(fd_set));
      FD_SET(sockfd,&f_r);	
      FD_SET(pttyfd,&f_r);
      s = select(128, &f_r, NULL, NULL, ptrtimer); 
      if (s == 0) 
         break;   /* timer expired: close the connection */
      if (s < 0) 
         {
         if (errno == EINTR)
            continue;  /* it's not a real error, is it a signal */
         fprintf(fd_log, "%schild processs %d: select error %d (s = %d)\n",GetTime(), getpid(), errno,s);
         fflush (fd_log);
         sleep (10);
         continue;
         }
      if (FD_ISSET(pttyfd, &f_r))
         {					/* car received from ptty */
         int p, w;

         r = read(pttyfd, buf, 30); 
         if (r < 0) 
            {
            fprintf(fd_log, "%schild process %d: can not read from %s (err %d)\n", GetTime(), getpid(), ttyname, errno);
            fflush (fd_log);
            break;
            }
         if (r == 0) 
            {
            fprintf(fd_log, "%schild process %d: read no byte from %s\n", GetTime (), getpid(), ttyname);
            fflush (fd_log);
            continue;
            }
         p = 0;
         totalout+=r;
         while (p < r)	/* I must write exactely the r characters */
            {
            w = write (sockfd, &buf[p], r-p);
            if (w < 0) 
               {
               fprintf(fd_log, "%schild process %d: can not write %d byte(s) on %s/%d (err %d)\n",GetTime(), getpid(), r-p,axname, tcpport, errno);
               fflush (fd_log);
               break;
               }
            p += w;
            }
         }
      if (FD_ISSET(sockfd, &f_r))
         {					/* car received from socket */
         int p, w;

         r = read(sockfd, buf, 30); 
         if (r < 0) 
            {
            if (errno == ETIMEDOUT)
               fprintf(fd_log, "%schild process %d: the Axel device %s is down.\n", GetTime(), getpid(), axname);
               else
               fprintf(fd_log, "%schild process %d: can not read from %s/%d (err %d)\n", GetTime(), getpid(), axname, tcpport, errno);
            fflush (fd_log);
            break;
            }
         totalin+=r;
         p = 0;
         while (p < r)	/* I must write exactely the r characters */
            {
            w = write (pttyfd, &buf[p], r-p);
            if (w < 0) 
               {
               fprintf(fd_log, "%schild process %d: can not write %d byte(s) to %s (err %d)\n", GetTime(), getpid(), r-p, ttyname, errno);
               fflush (fd_log);
               break;
               }
            p += w;
            }
         if ((rtty_maxchar_in !=0) && (totalin >= rtty_maxchar_in))
           break;
         }
      }
   fprintf(fd_log, "%sConnection %s/%d closed (%d in / %d out)\n\n", GetTime(), axname, tcpport, totalin, totalout);
   fflush (fd_log);
   if (do_flush) 
      {
      int nbflush;

      nbflush+=0; 
      while (1)
        {
        memset (&f_r, 0, sizeof(fd_set));
        FD_SET(slavefd,&f_r);
        timer.tv_sec=1;
        timer.tv_usec=0;
        ptrtimer=&timer;
        s = select(slavefd+1, &f_r, NULL, NULL, ptrtimer); 
        if (s == 0) 
           {
           break;   /* timer expired: flush over */
           }
        if (FD_ISSET(slavefd, &f_r))
           {					/* car received from ptty */
           r = read(slavefd, buf, 1); 
           nbflush+=r; 
           }
        }
     fprintf(fd_log, "%sFlushing %s: %d byte(s) dropped\n", GetTime(), ttyname, nbflush);
     fflush (fd_log);
     }
   close (sockfd);
   }
}



void ClientTTY ()
{
int i;
int childpid;
char str[30];

fprintf (fd_log, "\n%sStarting %d TTY Client(s):\n", GetTime(), nb_tty_clt);
fflush (fd_log);
for (i=0;i<nb_ax;i++)
   {
   if (tcpport[i] == 0)
      continue;
   if ((childpid = fork ()) < 0)
      {				/* I create a new child for this access */
      sprintf (str, "fork error = %d\r\n", errno);
      EndPrg (str);
      }
   if (childpid > 0)
      {	/* parent process */	
      fprintf (fd_log, "   - Process %d is waiting for characters from %s...\n", childpid, axtab[i][3]);
      fflush (fd_log);
      axpid[i]=childpid;
      }
   if (childpid == 0)
      {		/* child process */
      int j;

      /* close extra file descriptors */
      for (j=0;j<nb_ax;j++)
         {
         if (i != j)
            {
            close (sfd[j]);		/* slave fd */
            close (mfd[j]);		/* master fd */
            }
         }
      signal (SIGUSR1, SIG_IGN);	
      DoClientTTY (axtab[i][0], tcpport[i], mfd[i], axtab[i][3], sfd[i]);
      }
   }
}

int IdConn (fd, adr, aux)

int fd;
char *adr, *aux;
{
int lg, cpt, s;
char buf[MAXBUFSIZE];
char *ptr;
fd_set f_r;
struct timeval timer;
struct timeval *ptrtimer;

lg = 0;
cpt = 0;
while (1)
   {		/* identify the ax3000 (name and port expected */
   timer.tv_sec=1;
   timer.tv_usec=0;
   ptrtimer=&timer;
   memset (&f_r, 0, sizeof(fd_set));
   FD_SET(fd,&f_r);			/* READ set contains stdin */
   s = select(fd+1, &f_r, NULL, NULL, ptrtimer); 
   if (s == 0) 
      {   /* timer expired */
#if defined (UW7)
      size_t my_lg;
#else
#if defined (SCO6)
      socklen_t    my_lg;
#else
      int    my_lg;
#endif
#endif
      char my_addr[30];
      struct sockaddr_in ax_addr;

      my_lg = sizeof (ax_addr);
      if (getpeername (fd, (struct sockaddr *)&ax_addr, &my_lg) == -1)
         sprintf(my_addr, "unknown");
         else
         sprintf(my_addr, "%s",inet_ntoa(ax_addr.sin_addr));

      if (lg == 0)
         fprintf (fd_log, "%sIdstring from %s not received", GetTime(), my_addr);
         else
         fprintf (fd_log, "%sIdstring from %s too short", GetTime(), my_addr);
      return (-1);
      }
   if (s < 0) 
      if (errno == EINTR)
         continue;  /* it's not a real error, is it a signal */
   if (read (fd, &buf[lg], 1) <= 0)
      {
      fprintf (fd_log, "socket IdConn: read error = %d\n", errno);
      return (-1);
      }
   if (buf[lg] == 0)		/* is it the end of string ? */
      {
      cpt++;			/* yes */
      if (cpt == 2)		/* is it the second string ? */
         break;
      }
   lg++;
   if (lg > 25)	/* greater than xxx.xxx.xxx.xxx0parallel0 */
      {
      fprintf (fd_log, "%sBad Idstring", GetTime());
      return (-1);
      }
   }
strcpy (adr, buf);
ptr = strchr (buf, 0x00);
strcpy (aux, ptr+1);		/* OK, I get name and port */
return (0);
}

/*********************************************************/
/*                  M A I N                              */
/*********************************************************/
main(argc, argv)

int argc;
char **argv;
{
char str[MAX_LINE];
char logfile[MAX_LINE];
int i;
int sockfd, newsockfd;
#if defined (UW7)
size_t clilen;
#else
#if defined (SCO6)
      socklen_t    clilen;
#else
int    clilen;
#endif
#endif
int tcp_port;
struct sockaddr_in tty_addr, cli_addr;
void (*func) ();
pid_t childpid;
int ret, n;
int err;
char adr[30];
char aux[30];
int keep = 1;

father_pid = getpid();
strcpy (ax_file, AXFILE);
strcpy (logfile, LOGFILE);
tcp_port = TCPPORT;
NapSize = 0;
NapTime = 10;
do_flush=0;
do_unidirectional=0;
err = 0;
i = 1;
use_buff=0;
use_keepalive=1;
rtty_timer = RTTYTIMER; 
rtty_maxchar_in = RTTYMAXIN;
while (i<argc)
   {
   if (strcmp(argv[i],"-h") == 0)		/* display on-line help */
      {
      err = 1;
      break;
      }
   if (strcmp(argv[i],"-U") == 0)		/* do flush */
      {
      i++;
      do_unidirectional=1;
      continue;
      }
   if (strcmp(argv[i],"-F") == 0)		/* do flush */
      {
      i++;
      do_flush=1;
      continue;
      }
   if (strcmp(argv[i],"-b") == 0)		/* using buffer */
      {
      i++;
      use_buff=1;
      continue;
      }
   if (strcmp(argv[i],"-k") == 0)		/* disabling keepalive */
      {
      i++;
      use_keepalive=0;
      continue;
      }
   if (strcmp(argv[i],"-f") == 0)		/* association file */
      {
      i++;
      strcpy (ax_file, argv[i]);	
      i++;
      continue;
      }
   if (strcmp(argv[i],"-l") == 0)		/* log file */
      {
      i++;
      strcpy (logfile, argv[i]);	
      i++;
      continue;
      }
   if (strcmp(argv[i],"-rmi") == 0)		/* RTTY Max char in  */
      {
      i++;
      rtty_maxchar_in = atoi(argv[i]);	
      i++;
      if (rtty_maxchar_in < 0)
         {
         err = 1;
         break;
         }
      continue;
      }
   if (strcmp(argv[i],"-rt") == 0)		/* RTTY Timer  */
      {
      i++;
      rtty_timer = atoi(argv[i]);	
      i++;
      if (rtty_timer < 0)
         {
         err = 1;
         break;
         }
      continue;
      }
   if (strcmp(argv[i],"-ns") == 0)             /* napsize characters */
      {
      i++;
      NapSize = atoi(argv[i]);    
      i++;
      continue;
      }
   if (strcmp(argv[i],"-nt") == 0)       /* naptime delay (micro sec)*/
      {
      i++;
      NapTime = atoi(argv[i]);    
      i++;
      continue;
      }
   if (strcmp(argv[i],"-n") == 0)		/* TCP Port number */
      {
      i++;
      tcp_port = atoi(argv[i]);	
      i++;
      if (tcp_port <= 0)
         {
         err = 1;
         break;
         }
      continue;
      }
   err = 1;
   break;
   }

if (err)			/* pb with command line */
   {
   fprintf(stderr,"AXEL remote tty server for %s\n", OS);
   fprintf(stderr,"Version %s - (C) 1996-2007 - AXEL\n", VERSION);
#ifdef CHECK_OPEN_SLAVE
fprintf(stderr, "IMPORTANT : CHECK MODE ENABLED\n");
#endif
#ifdef PTMX
   fprintf(stderr,"PTMX/PTC devices supported\n");
#endif
   fprintf(stderr,"\nUsage: \n");

   fprintf(stderr,"  %s [-hUFkb] [-f file] [-l file] [-n port] [-rmi nb] -[rt sec] &\n", argv[0]);
   fprintf(stderr,"     -f file : associations file (device-port <-> ptty)\n");
   fprintf(stderr,"               default: %s\n", AXFILE);
   fprintf(stderr,"     -l file : log file (to store events)\n");
   fprintf(stderr,"               default: %s\n", LOGFILE);
   fprintf(stderr,"     -n port : TCP Port number\n");
   fprintf(stderr,"               default: %d\n", TCPPORT);
   fprintf(stderr,"     -b      : using a buffer for data from the socket\n");
   fprintf(stderr,"               default: no buffer\n");
   fprintf(stderr,"     -F      : flush ptty when a connection is opened\n");
   fprintf(stderr,"               default: no flush\n");
   fprintf(stderr,"     -U      : uni-directional mode (UNIX->Axel device)\n");
   fprintf(stderr,"               default: bi-directional\n");
   fprintf(stderr,"     -k      : turn off keepalive function\n");
   fprintf(stderr,"               default: keepalive function enabled\n");
   fprintf(stderr,"     -rt sec : rtty inactivity timer (in seconds). 0 is no timer.\n");
   fprintf(stderr,"               default: %d\n", RTTYTIMER);
   fprintf(stderr,"     -rmi nb : rtty max incoming char. 0 is no max.\n");
   fprintf(stderr,"               default: %d\n", RTTYMAXIN);
   fprintf(stderr,"     -ns nbr : to avoid CPU overload, axttyd is asleep after writing -ns bytes\n");
   fprintf(stderr,"               default: 0 (no optimization)\n");
   fprintf(stderr,"     -nt msec: delay in msec during axttyd is asleep (see -ns)\n");
   fprintf(stderr,"               default: 10\n");
   fprintf(stderr,"     -h      : display on-line help\n");
   exit(1);
   }

if ((fd_log = fopen (logfile, "w")) == NULL)
   {
   fprintf(stderr,"Can not open %s!\n", logfile);
   exit (1);
   }

fprintf(fd_log, "\nAXEL remote tty Server (%s) V%s : pid %d\n", OS, VERSION, getpid());

#ifdef CHECK_OPEN_SLAVE
fprintf(fd_log, "\nIMPORTANT : CHECK MODE ENABLED\n\n");
#endif

fprintf(fd_log, "PTMX/PTC devices:      ");
#ifdef PTMX
   fprintf(fd_log, "supported\n");
#else
   fprintf(fd_log, "not supported\n");
#endif
fprintf(fd_log, "TCP Port number:       %d\n", tcp_port);
fprintf(fd_log, "Data buffer:           ");
if (use_buff == 0)
   fprintf(fd_log, "no\n");
   else
   fprintf(fd_log, "%d bytes (high water mark: %d)\n", MAXBUFFSOCK, HIGHWATERMARK);
fprintf(fd_log, "Ptty flush:            ");
if (do_flush == 0)
   fprintf(fd_log, "no\n");
   else
   fprintf(fd_log, "yes\n");

fprintf(fd_log, "Communication:         ");
if (do_unidirectional == 0)
   fprintf(fd_log, "bi-directional\n");
   else
   fprintf(fd_log, "uni-directional\n");

fprintf(fd_log, "CPU saver:             ");
if (NapSize == 0)
   fprintf(fd_log, "no\n");
   else
   fprintf(fd_log, "yes (max char: %d, delay: %d msec)\n", NapSize, NapTime);

#if ! defined (SCO5) && ! defined (SCO3)
NapTime *= 1000;			/* milli-second conversion */
#endif

fprintf(fd_log, "Keepalive:             ");
if (use_keepalive == 1)
   fprintf(fd_log, "enabled\n");
   else
   fprintf(fd_log, "disabled\n");

fprintf(fd_log, "RTTY max. incoming:    ");
if (rtty_maxchar_in == 0)
   fprintf(fd_log, "disabled\n");
   else
   fprintf(fd_log, "%d bytes\n", rtty_maxchar_in);

fprintf(fd_log, "RTTY inactivity timer: ");
if (rtty_timer == 0)
   fprintf(fd_log, "disabled\n");
   else
   fprintf(fd_log, "%d seconds\n", rtty_timer);

for (i=0;i<MAX_ENTRY;i++)
   {			/* initialize association array */
   mfd[i] = -1;
   sfd[i] = -1;
   axpid[i] = -1;
   mark[i] = 0;
   first[i] = 1;
   }

nb_ax = ParseFile (axtab, ax_file);	/* examine association file */

close (0);				/* close not used file descriptors */
close (1);
close (2);

if (nb_ax > 0)				/* if one valid entry at least */
   AllocatePtty ();			/* try to open ptyp and ttyp */


/* socket initialisations */
fprintf (fd_log, "\n%sInitialisations in progress...\n", GetTime());
if ((sockfd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
   EndPrg ("Can not open stream socket");

if (setsockopt (sockfd, SOL_SOCKET, SO_KEEPALIVE, (void *)&use_keepalive, sizeof(use_keepalive)) < 0)
      {
      char str[60];

      sprintf(str, "Can not set keep alive to %d (error %d)", use_keepalive, errno);
      EndPrg (str);
      }

while (1)
   {
   memset ((char *) &tty_addr, 0, sizeof (tty_addr));
   tty_addr.sin_family = AF_INET;
   tty_addr.sin_addr.s_addr = htonl (INADDR_ANY);
   tty_addr.sin_port = htons (tcp_port);
   #ifdef ICL
   if (bind (sockfd, (caddr_t) &tty_addr, sizeof (tty_addr)) < 0)
   #else
   if (bind (sockfd, (struct sockaddr *) &tty_addr, sizeof (tty_addr)) < 0)
   #endif
      {
      char str[60];

      fprintf(fd_log, "%scan not bind TCP port (err %d). Retry in 5 seconds...\n", GetTime(), errno);
      fflush (fd_log);
      sleep (5);
      }
      else
      break;
   }

func = CleanAll;			/* to clean at the end of program */
signal (SIGHUP, func);	
signal (SIGTERM, func);

func = EndChild;
signal (SIGCLD, func);			/* to avoid zombies */

func = ReExamineFile;			/* to re-examine association file */
signal (SIGUSR1, func);		

if (listen (sockfd, MAX_CONN) < 0)		/* max MAX_CONN pending */
   {
   char str[60];

   sprintf(str, "Can not listen (error %d)", errno);
   EndPrg (str);
   }

#ifdef LYNX
setpgrp (0, getpid());
#else
setpgrp ();
#endif

if (nb_tty_clt > 0)
   ClientTTY (); 

/* main loop */
fprintf (fd_log, "\n%sWaiting for connections from TCP/IP socket...\n\n", GetTime());
fflush (fd_log);

while (1) 
   {
   clilen = sizeof (cli_addr);
#ifdef ICL
   newsockfd = accept (sockfd, (caddr_t) &cli_addr, &clilen);
#else
   newsockfd = accept (sockfd, (struct sockaddr *) &cli_addr, &clilen);
#endif
/* a new connection is asked */
   if (newsockfd < 0)
      {
      if (errno == EINTR)
         {
         errno = 0;			/* reset errno */
         sprintf (str, "accept error: reset\n");
         continue;			/* eventually a child exited */
         }
      sprintf (str, "accept error = %d\n", errno);
      EndPrg (str);
      }
   if (IdConn (newsockfd, adr, aux) == -1) 
      {			/* not received the 2 strings */
      fprintf (fd_log, " - Closed!\n");
      fflush (fd_log);
      shutdown (newsockfd, 2);
      close (newsockfd);
      continue;		/* ignore connection attempt */
      }
   ret = -1;
   for (i=0;i<nb_ax;i++)
      if ((strcmp (axtab[i][0], adr) == 0) && (strcmp (axtab[i][1], aux) == 0))
         {
         ret = i;	/* the name and port match with this entry */
         break;
         }
   if (ret == -1)
      {			/* no entry matches */
      fprintf(fd_log, "%sConnection from '%s %s': access denied.\n", GetTime (), adr, aux);
      fflush (fd_log);
      shutdown (newsockfd, 2);
      close (newsockfd);
      continue;
      }
   if (axpid[ret] != -1)	/* is it an existing child for this acces ? */
      {
      fprintf(fd_log, "%sI must kill a former connection (pid: %d)\n", GetTime(),axpid[ret]);
      fflush (fd_log);
      kill (axpid[ret], 9);
      }
#ifdef CHECK_OPEN_SLAVE
/* the master was maybe closed by a former connection. So we have to */
/* open it. This is a normal process */
   if (mfd[ret] == -1)
      {
      mfd[ret]=OpenPTTY (axtab[ret][2]);
      fprintf(fd_log, "%sCHECK MODE: father re-open %s for a new connection\n", GetTime(),axtab[ret][2]);
      fflush (fd_log);
      }
#endif
   if (mfd[ret] == -1)
      {			/* no ptty associated */
      fprintf(fd_log, "%sConnection from '%s %s': rejected (%s not opened).\n", GetTime (), adr, aux, axtab[ret][2]);
      fflush (fd_log);
      shutdown (newsockfd, 2);
      close (newsockfd);
      continue;
      }
   fprintf(fd_log, "%sConnection from '%s %s': accepted.\n", GetTime(), adr, aux);
   fflush (fd_log);
   if ((childpid = fork ()) < 0)
      {				/* I create a new child for this access */
      sprintf (str, "fork error = %d\r\n", errno);
      EndPrg (str);
      }
   if (childpid == 0)
      {		/* child process */
      /* close extra file descriptors */
      close (sockfd);			/* listen socket */
      for (i=0;i<nb_ax;i++)
         {
         close (sfd[i]);		/* slave fd */
         if (i != ret)
            close (mfd[i]);		/* master fd */
         }
      signal (SIGUSR1, SIG_IGN);	
      PerformConn(newsockfd, mfd[ret], first[ret], axtab[ret][2]);	/* child is in charge of this access  */
      }
   fprintf(fd_log, "%sINFO: the process %d controls '%s %s'.\n", GetTime(), childpid, adr, aux);
   fflush (fd_log);
   axpid[ret] = childpid;
   first[ret] = 0;
   close (newsockfd);		       /* parent waits new connections */
#ifdef CHECK_OPEN_SLAVE
   /* the master fd must be closed by the father. This allows the son to   */
   /* close and open the master ptty every time is needed */
   close (mfd[ret]);
   mfd[ret]=-1;
   fprintf(fd_log, "%sCHECK MODE: father closes %s.\n", GetTime(), axtab[ret][2]);
   fflush (fd_log);
#endif
   }
}
